/*
Copyright (c) 2011, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <grub/extcmd.h>
#include <grub/types.h>
#include <grub/misc.h>
#include <grub/mm.h>
#include <grub/env.h>
#include <grub/err.h>
#include <grub/dl.h>
#include <grub/term.h>
#include <grub/time.h>
#include "c.h"
#include "datatype.h"
#include "bitsutil.h"
#include "smp.h"

GRUB_MOD_LICENSE("GPLv3+");
GRUB_MOD_DUAL_LICENSE("3-clause BSD");

enum verbose_level {
    V_NONE,
    V_FAIL,
    V_DETAIL,
    V_PASS,
};

static enum verbose_level verbose = V_DETAIL;
static U32 ncpus = 0;
static const CPU_INFO *cpu;
#define ALL_CPUS (~0UL)

struct mem {
    U32 addr;
    U32 value;
};

struct unique_mem {
    U32 count;
    struct mem mem;
};

static struct mem *mem = 0;
static struct unique_mem *unique_mem = 0;

static U32 unique_count;

static grub_err_t parse_cpu_num(char *str, U32 * out)
{
    U32 num;

    grub_errno = GRUB_ERR_NONE;

    if (grub_strcmp(str, "all") == 0) {
        *out = ALL_CPUS;
        return GRUB_ERR_NONE;
    }

    if (strtou32_h(str, &num) != GRUB_ERR_NONE)
        return grub_errno;

    if (num >= ncpus)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "CPU number %u invalid", num);

    *out = num;
    return GRUB_ERR_NONE;
}

static U32 viewpoint_env(void)
{
    U32 num;
    const char *env = grub_env_get("viewpoint");
    if (!env)
        return ALL_CPUS;
    if (strtou32_h(env, &num) != GRUB_ERR_NONE || num >= ncpus) {
        grub_printf("viewpoint environment variable contains invalid value \"%s\"\n", env);
        return ALL_CPUS;
    }
    return num;
}

static U32 first_cpu(U32 cpu_num)
{
    return cpu_num == ALL_CPUS ? 0 : cpu_num;
}

static U32 last_cpu(U32 cpu_num)
{
    return cpu_num == ALL_CPUS ? ncpus : cpu_num + 1;
}

static void uninit(void)
{
    grub_free(mem);
    grub_free(unique_mem);
}

static grub_err_t oom_uninit(void)
{
    uninit();
    return grub_error(GRUB_ERR_OUT_OF_MEMORY, "Out of memory");
}

static grub_err_t init(void)
{
    uninit();

    ncpus = smp_init();
    if (!ncpus)
        return grub_error(GRUB_ERR_IO, "Failed to initialize SMP");

    cpu = smp_read_cpu_list();
    if (!cpu)
        return grub_error(GRUB_ERR_IO, "Failed to initialize SMP (smp_read_cpu_list)");

    mem = grub_malloc(ncpus * sizeof(*mem));
    if (!mem)
        return oom_uninit();

    unique_mem = grub_malloc(ncpus * sizeof(*unique_mem));
    if (!unique_mem)
        return oom_uninit();

    return GRUB_ERR_NONE;
}

static const char c_help[] =
    "Usage: c \"C-style expression with space-separated tokens\"\n"
    "Evaluate a C expression\n"
    "All evaluation occurs on 64-bit unsigned integers, specified as decimal\n"
    "numbers, hex numbers prefixed by 0x, or named variables taken from the\n"
    "environment.  Variable assignments work, and set variables in the environment.\n"
    "Returns true or false based on the final value of the expression, so\n"
    "\"if c ... \" and \"while c ...\" work.\n"
    "\n"
    "Supported operators, in order from highest to lowest precedence:\n"
    " ()        (parentheses)\n"
    " ! ~ ++ -- (both pre- and post- increment and decrement)\n"
    " * / %\n"
    " + -\n"
    " << >>\n"
    " < <= > >=\n"
    " == !=\n"
    " &         (bitwise and)\n"
    " ^         (bitwise xor)\n"
    " |         (bitwise or)\n"
    " &&        (logical and; WARNING: does not short-circuit)\n"
    " ||        (logical or;  WARNING: does not short-circuit)\n"
    " = += -= *= /= %= <<= >>= &= ^= |=\n"
    " ,         (comma operator)\n";

static grub_err_t grub_cmd_c(grub_command_t cmd, int argc, char **args)
{
    U64 result;
    (void)cmd;

    if (argc == 1 && grub_strcmp(args[0], "--help") == 0) {
        grub_printf("%s", c_help);
        return GRUB_ERR_NONE;
    }

    if (c_expr(argc, args, &result))
        return !result;
    else
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Failed to parse C expression");
}

static void noop_callback(void *param)
{
    (void)param;
}

static const struct grub_arg_option cpu_ping_options[] = {
#undef OPTION_CPU
#define OPTION_CPU 0
    {"cpu", 'c', 0, "CPU number", "CPU", ARG_TYPE_STRING},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_cpu_ping(struct grub_extcmd_context *context, int argc, char **args)
{
    U32 i, j;
    U32 cpuNum;
    U32 repeat_count;
    U64 start, stop;
    U32 seconds = 0;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    cpuNum = viewpoint_env();

    if (context->state[OPTION_CPU].set)
        if (parse_cpu_num(context->state[OPTION_CPU].arg, &cpuNum) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: repeat_count");
    if (strtou32_h(args[0], &repeat_count) != GRUB_ERR_NONE)
        return grub_errno;

    start = grub_get_time_ms();
    for (j = 0; j < repeat_count; j++) {
        if (grub_getkey_noblock() == GRUB_TERM_ESC)
            break;
        stop = grub_get_time_ms();
        if (stop - start > 1000) {
            start = stop;
            seconds++;
            grub_printf("\r%u second%s (%u%%)", seconds, seconds == 1 ? "" : "s", (j * 100) / repeat_count);
        }
        for (i = first_cpu(cpuNum); i != last_cpu(cpuNum); i++)
            smp_function(cpu[i].apicid, noop_callback, NULL);
    }
    grub_printf("\r");

    return GRUB_ERR_NONE;
}

static void mem_callback(void *param)
{
    struct mem *m = param;
    m->value = *(U32 *)m->addr;
}

static void get_mem(U32 cpu_num, U32 addr, U64 mask, U32 shift)
{
    U32 i, j;

    unique_count = 0;

    for (i = first_cpu(cpu_num); i != last_cpu(cpu_num); i++) {
        bool found = false;
        mem[i].addr = addr;
        if (!smp_function(cpu[i].apicid, mem_callback, &mem[i]))
            grub_printf("DEBUG: smp_function failed in get_mem\n");
        mem[i].value >>= shift;
        mem[i].value &= mask;

        for (j = 0; j < unique_count; j++)
            if (unique_mem[j].mem.value == mem[i].value) {
                unique_mem[j].count++;
                found = true;
                break;
            }
        if (!found) {
            unique_mem[unique_count].count = 1;
            unique_mem[unique_count].mem = mem[i];
            unique_count++;
        }
    }
}

static const struct grub_arg_option rdmem_options[] = {
#undef OPTION_CPU
#define OPTION_CPU 0
    {"cpu", 'c', 0, "CPU number", "CPU", ARG_TYPE_STRING},
#undef OPTION_VARNAME
#define OPTION_VARNAME 1
    {"varname", 'v', 0, "Save read value into variable VARNAME", "VARNAME", ARG_TYPE_STRING},
#undef OPTION_MASK
#define OPTION_MASK 2
    {"mask", 'm', 0, "Mask to apply to value (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_SHIFT
#define OPTION_SHIFT 3
    {"shift", 's', 0, "Shift count for mask and value (default=0)", "SHIFT", ARG_TYPE_STRING},
#undef OPTION_QUIET
#define OPTION_QUIET 4
    {"quiet", 'q', 0, "No display unless failures occur (default=disabled)", 0, 0},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_rdmem(struct grub_extcmd_context *context, int argc, char **args)
{
    U32 cpuNum;
    U32 memAddr;
    const char *varname;
    U64 mask = ~0ULL;
    U32 shift = 0;
    U32 i, j;
    char mem_desc[128];
    bool status = true;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    cpuNum = viewpoint_env();

    if (context->state[OPTION_CPU].set)
        if (parse_cpu_num(context->state[OPTION_CPU].arg, &cpuNum) != GRUB_ERR_NONE)
            return grub_errno;

    if (context->state[OPTION_MASK].set)
        if (strtou64_h(context->state[OPTION_MASK].arg, &mask) != GRUB_ERR_NONE)
            return grub_errno;

    if (context->state[OPTION_SHIFT].set)
        if (strtou32_h(context->state[OPTION_SHIFT].arg, &shift) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: memory_address");
    if (strtou32_h(args[0], &memAddr) != GRUB_ERR_NONE)
        return grub_errno;

    if (context->state[OPTION_VARNAME].set)
        varname = context->state[OPTION_VARNAME].arg;
    else
        varname = NULL;

    get_mem(cpuNum, memAddr, mask, shift);

    // Set fail status for more than one unique value
    if (unique_count != 1)
        status = false;

    if (!context->state[OPTION_QUIET].set)
        for (i = 0; i < unique_count; i++) {
            if (mask == ~0ULL) {
                if (shift == 0)
                    grub_snprintf(mem_desc, sizeof(mem_desc), "memory 0x%x", memAddr);
                else
                    grub_snprintf(mem_desc, sizeof(mem_desc), "memory 0x%x shr %u", memAddr, shift);
            } else {
                if (shift == 0)
                    grub_snprintf(mem_desc, sizeof(mem_desc), "memory 0x%x & 0x%llx", memAddr, mask);
                else
                    grub_snprintf(mem_desc, sizeof(mem_desc), "memory 0x%x shr %u & 0x%llx", memAddr, shift, mask);
            }

            grub_printf("%s = 0x%x\n", mem_desc, unique_mem[i].mem.value);

            grub_printf("On %u CPUs:%s", unique_mem[i].count, unique_mem[i].count == ncpus ? " all" : "");
            if (unique_mem[i].count != ncpus)
                for (j = first_cpu(cpuNum); j < last_cpu(cpuNum); j++)
                    if (unique_mem[i].mem.value == mem[j].value)
                        grub_printf(" 0x%x", cpu[j].apicid);
            grub_printf("\n");
        }

    if (varname) {
        char buf[sizeof("0xXXXXXXXXXXXXXXXX")];
        if (unique_count != 1) {
            if (!context->state[OPTION_QUIET].set)
                grub_printf("Variable setting requested but memory value not unique\n");
            return grub_error(GRUB_ERR_TEST_FAILURE, "false");
        }
        grub_snprintf(buf, sizeof(buf), "0x%x", unique_mem[0].mem.value);
        grub_env_set(varname, buf);
    }

    return status ? GRUB_ERR_NONE : grub_error(GRUB_ERR_TEST_FAILURE, "false");
}

static const char *
grub_env_read_test_verbose (struct grub_env_var *var __attribute__ ((unused)), const char *val __attribute__ ((unused)))
{
    // Ignore the string value; return the current value of verbose
    static char buf[sizeof("4294967296")];
    grub_snprintf(buf, sizeof(buf), "%u", verbose);
    return buf;
}

static char *
grub_env_write_test_verbose (struct grub_env_var *var __attribute__ ((unused)), const char *val)
{
    if (strtou32_h(val, &verbose) != GRUB_ERR_NONE)
        return 0;

    return grub_strdup ("");
}

static grub_command_t cmd_c;
static grub_extcmd_t cmd_cpu_ping;
static grub_extcmd_t cmd_rdmem;

GRUB_MOD_INIT(testsuite)
{
    // Export these environment variables to make them available in all contexts
    grub_env_export ("test_verbose");

    // Register a environment write hook
    grub_register_variable_hook ("test_verbose", grub_env_read_test_verbose, grub_env_write_test_verbose);

  cmd_c = grub_register_command("c", grub_cmd_c, "\"C-style expression with space-separated tokens\"", "Evaluate a C expression.");
  cmd_cpu_ping = grub_register_extcmd("cpu_ping", grub_cmd_cpu_ping, 0,
                                      "[-c cpu_num] count",
                                      "Ping CPU",
                                      cpu_ping_options);
  cmd_rdmem = grub_register_extcmd("rdmem", grub_cmd_rdmem, 0,
                                   "[-q] [-c cpu_num] [-m mask] [-v VARNAME] [-s shift] memory_addr",
                                   "Read memory",
                                   rdmem_options);
}

GRUB_MOD_FINI(testsuite)
{
    grub_unregister_extcmd(cmd_rdmem);
    grub_unregister_extcmd(cmd_cpu_ping);
    grub_unregister_command(cmd_c);

    // Clear the environment hooks
    grub_register_variable_hook ("test_verbose", 0, 0);

    uninit();
}
